import asyncio
import logging
from logging import Logger
import uuid
import sys
from types import FunctionType
from typing import Union

from aio_pika import Message, DeliveryMode, IncomingMessage

from . import RabbitMQConnectionData, AsyncConnection, Consumer, Publisher


class AsyncAdapter:
    """
    An asynchronous RabbitMQ client which is fully integrated with aio-pika.
    All-in-one adapter which provides the following interfaces:
    1. publish - Publish a message.
    2. receive - Consume messages from a queue. Can automatically reply to desired routes if the received message
                 contains a "reply_to" property.
    3. rpc - Publish a message with a reply_to property, wait for a reply message and return the reply's value.

    Architecture:
    This class uses a single connection, and creates a Consumer for each receive channel passed in the configurations,
    and a Publisher for each publish channel passed in the configuration.
    """

    _rabbitmq_connection_data: RabbitMQConnectionData
    _logger: Logger
    _loop: asyncio.AbstractEventLoop
    _configuration: dict

    def __init__(self, rabbitmq_connection_data: RabbitMQConnectionData, configuration: dict, logger: Logger = None,
                 loop: asyncio.AbstractEventLoop = None, connection_properties: dict = None,
                 connection_timeout: Union[int, float] = 10, connection_attempts: int = 5, attempt_backoff: int = 5):
        """
        :param rabbitmq_connection_data: Connection data for RabbitMQ server. Used to open AsyncConnection.
        :param configuration: RabbitMQ configuration for both consuming and publishing.
                              It is separated by the keys `receive` and `publish`
        :param logger: Logger to use. If None, creates a default Logger.
        :param loop: Async event loop. If it is None, defaults to asyncio.get_event_loop()
        :param connection_properties: Connection properties
        :param connection_timeout: Timeout of each connection attempt
        :param connection_attempts: Number of attempts for connection
        :param attempt_backoff: Seconds waited between connection attempts
        """
        self._rabbitmq_connection_data = rabbitmq_connection_data
        self._logger = logger or self._create_logger()
        self._loop = loop or asyncio.get_event_loop()
        self._configuration = configuration

        self._connection = AsyncConnection(
            rabbitmq_connection_data,
            self.logger,
            self._loop,
            connection_properties,
            connection_timeout,
            connection_attempts,
            attempt_backoff
        )

        self._rpc_corr_id_dict = dict()

        self._default_publisher = Publisher(connection=self._connection,
                                            logger=self.logger,
                                            exchange_name="",
                                            loop=self._loop)

        self._publishers = dict()
        for publish_configuration in self._configuration.get("publish", dict()).values():
            self.add_publisher(publish_configuration)

        self._consumers = dict()
        for receive_configuration in self._configuration.get("receive", dict()).values():
            self.add_consumer(receive_configuration)

    @staticmethod
    def _create_logger() -> Logger:
        """
        Creates logger with hardcoded handlers and formatters
        :return: Created logger
        """
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.DEBUG)
        sh = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter('[%(asctime)s] [%(filename)-25s] [%(levelname)-8s] %(message)s')
        sh.setFormatter(formatter)
        sh.setLevel(logging.DEBUG)
        logger.addHandler(sh)
        logger.propagate = False
        return logger

    @property
    def logger(self) -> Logger:
        """
        :return: self._logger
        """
        return self._logger

    def add_publisher(self, configuration: dict) -> Publisher:
        """
        Creates a Publisher with the given configurations and add to self._publishers
        :param configuration: Configuration to pass to publisher
        :return: Created publisher
        """
        publisher = Publisher(connection=self._connection, logger=self.logger, loop=self._loop, **configuration)
        self._publishers[configuration["exchange_name"]] = publisher
        return publisher

    def add_consumer(self, configuration: dict) -> Consumer:
        """
        Creates a Consumer with the given configurations and add to self._consumers
        :param configuration: Configurations to pass to Consumer
        :return: Created consumer
        """
        consumer = Consumer(connection=self._connection, logger=self.logger, loop=self._loop, **configuration)
        self._consumers[configuration["queue_name"]] = consumer
        return consumer

    async def publish(self, body: bytes, exchange: str, routing_key: str = None, properties: dict = None,
                      mandatory: bool = True, immediate: bool = False, timeout: Union[int, float, None] = None):
        """
        Publish a message using the publisher relevant to the exchange given.
        :param body: message
        :param exchange: The exchange to publish to
        :param routing_key: The routing key to send the message with
        :param properties: RabbitMQ message properties
        :param mandatory: RabbitMQ publish mandatory param
        :param immediate: Whether or not message should be immediate
        :param timeout: Publish timeout
        """
        self.logger.info("Trying to publish message")
        if properties is None:
            properties = dict(delivery_mode=DeliveryMode.PERSISTENT)

        publisher = self._publishers.get(exchange)
        if publisher is None:
            raise KeyError(f"There is no publisher for the given exchange: {exchange}")

        try:
            message = Message(body, **properties)
            await publisher.publish(
                message, routing_key=routing_key, mandatory=mandatory, immediate=immediate, timeout=timeout)
        except Exception:
            self.logger.exception("Failed to publish message")
            raise

    async def receive(self, handler, queue: str, no_ack: bool = False) -> None:
        """
        Consume messages with the consumer relevant to the queue given.
        If the received message contains a "reply_to" parameter, reply to the route with the message handler's result.
        :param handler: Message handler
        :type handler: async def fn(logger, incoming_message)
        :param queue: The queue to consume from
        :param no_ack: Whether or not to ack incoming messages
        """
        consumer = self._consumers.get(queue)
        if consumer is None:
            raise KeyError(f"There is no consumer for the given queue: {queue}")

        try:
            await consumer.consume(self._on_message, handler=handler, no_ack=no_ack)
        except Exception:
            self.logger.exception("Failed to receive message.")
            raise

    async def _on_message(self, message: IncomingMessage, handler: FunctionType) -> None:
        """
        Base message handler. Runs the given message handler on the received message, and if we want to reply to a
        given route, sends the handler's result as a reply.
        :param message: Consumed message to handle
        :param handler: Message handler
        """
        self.logger.info("Received a new message")
        try:
            result = await handler(self.logger, message)
            self.logger.info("Message has been processed successfully")
            if message.reply_to is not None:
                self.logger.info(f"Sending result back to "
                                 f"queue: {message.reply_to}, correlation id: {message.correlation_id}")
                response_message = Message(body=result,
                                           correlation_id=message.correlation_id,
                                           reply_to=message.reply_to)
                await self._default_publisher.default_exchange_publish(message=response_message,
                                                                       routing_key=message.reply_to,
                                                                       mandatory=False)
                self.logger.info(f"Sent result back to caller. "
                                 f"Queue: {message.reply_to}, correlation id: {message.correlation_id}")
        except Exception:
            self.logger.exception("Failed to handle received message.")
            raise
        finally:
            message.ack()

    async def rpc(self, body: bytes, receive_queue: str, publish_exchange: str, timeout: Union[int, float],
                  ttl: int) -> bytes:
        """
        RPC call. Consumes from the given receive_queue to wait for the result of the RPC call.
        Then publishes a message to the given publish_exchange with correlation_id and reply_to properties.
        Then waits for the RPC result until it is received or a timeout occurs.
        If a result arrives, returns it.
        :param body: Message body
        :param receive_queue: The queue to consume from
        :param publish_exchange: The exchange to publish to
        :param timeout: RPC timeout (seconds)
        :param ttl: Message's time to live in the RabbitMQ queue (seconds)
        :return: RPC call result
        :raises: Exception contained in future if any
        """
        consumer = self._consumers.get(receive_queue)
        if consumer is None:
            raise KeyError(f"There is no consumer for the given queue: {receive_queue}")

        self.logger.info(f"Preparing to rpc call. Publish exchange: {publish_exchange}; Receive queue: {receive_queue}")
        await consumer.consume(self._rpc_response_callback)

        correlation_id = self._prepare_rpc_correlation_id()
        properties = dict(correlation_id=correlation_id, reply_to=receive_queue, expiration=ttl * 1000)
        await self.publish(body, publish_exchange, properties=properties, mandatory=True)
        self.logger.info(f"RPC message has been sent. {correlation_id}")

        future = await self._wait_result(correlation_id, timeout)
        self.logger.info(f"RPC message gets response. {correlation_id}")
        if future.exception():
            self.logger.error(f"RPC future returned exception: {future.exception()}")
            raise future.exception()
        if correlation_id in self._rpc_corr_id_dict:
            del self._rpc_corr_id_dict[correlation_id]
        return future.result()

    def _prepare_rpc_correlation_id(self) -> str:
        """
        Prepares a valid correlation_id for an RPC call, ensuring that it doesn't already exist in
        self._rpc_corr_id_dict.
        :return: Correlation ID
        """
        correlation_id = str(uuid.uuid1())
        while correlation_id in self._rpc_corr_id_dict:
            correlation_id = str(uuid.uuid1())
        self.logger.info(f"Starting rpc calling correlation id: {correlation_id}")

        future = self._loop.create_future()
        self._rpc_corr_id_dict[correlation_id] = future
        return correlation_id

    def _rpc_response_callback(self, message: IncomingMessage) -> None:
        """
        Handles RPC response message, setting Future result to message body
        :param message: RPC response message
        """
        self.logger.info(f"Received RPC response. Correlation id: {message.correlation_id}")
        if message.correlation_id in self._rpc_corr_id_dict:
            self.logger.info(f"Received RPC response. Correlation id: {message.correlation_id} was found")
            self._rpc_corr_id_dict[message.correlation_id].set_result(message.body)
        else:
            self.logger.warning(f"Received unexpected RPC response. Correlation id: {message.correlation_id}")
        message.ack()

    async def _wait_result(self, corr_id: str, timeout: Union[int, float, None] = None) -> asyncio.Future:
        """
        Wait for RPC result for given correlation ID until the given timeout fires.
        :param corr_id: Correlation ID to wait for
        :param timeout: Timeout in seconds
        :return: Future whose result we set
        """
        self.logger.info(f"Starting to wait for result. {corr_id}")
        try:
            future = self._rpc_corr_id_dict[corr_id]
        except KeyError:
            future = asyncio.Future()
            future.set_exception(KeyError(f"No future exists for correlation ID {corr_id}"))
            return future

        try:
            await asyncio.wait_for(future, timeout=timeout)
        except asyncio.TimeoutError:
            self.logger.error(f"RPC timeout. Correlation id: {corr_id}")
            del self._rpc_corr_id_dict[corr_id]
            future = asyncio.Future()
            future.set_exception(TimeoutError(f'RPC timeout. Correlation id: {corr_id}'))

        return future

    def status_check(self) -> bool:
        """
        :return: Whether or not the RabbitMQ connection being used is connected
        """
        return self._connection.is_connected()



import asyncio

import pytest
from aio_pika import Message, IncomingMessage

from tests.conftest import collect_future
from tornado_bunny import Consumer, Publisher


class TestConsumerPublisher:
    _test_future: asyncio.Future = None

    @pytest.fixture(scope="function")
    def message(self) -> Message:
        return Message(b"test message")

    @pytest.mark.asyncio
    async def test_publish(self, message: Message, publisher: Publisher, consumer: Consumer) -> None:
        # Act
        await publisher._prepare_publish()
        queue = await publisher.channel_config.declare_queue("publish_test",
                                                             publisher._exchange,
                                                             publisher._routing_key,
                                                             publisher._durable,
                                                             publisher._auto_delete)
        message_count_pre = queue.declaration_result.message_count
        await publisher.publish(message=message)
        await queue.declare()
        message_count_post = queue.declaration_result.message_count

        # Assert
        assert message_count_post == message_count_pre + 1

        # Teardown
        await queue.delete(if_unused=False, if_empty=False)

    @pytest.mark.asyncio
    async def test_consume_publish(self,
                                   event_loop: asyncio.AbstractEventLoop,
                                   message: Message,
                                   publisher: Publisher,
                                   consumer: Consumer):
        # Arrange
        self._test_future = event_loop.create_future()

        # Act
        await consumer.consume(self.message_handler)
        await publisher.publish(message=message)
        message_body = await collect_future(self._test_future, timeout=10)

        # Assert
        assert consumer.channel_config.started
        assert message_body == message.body

    @pytest.mark.asyncio
    async def test_consume_publish_close_channel(self,
                                                 event_loop: asyncio.AbstractEventLoop,
                                                 message: Message,
                                                 publisher: Publisher,
                                                 consumer: Consumer):
        # Arrange
        self._test_future = event_loop.create_future()

        # Act
        await consumer.consume(self.message_handler)
        await consumer.channel_config.close_channel(None)
        await asyncio.sleep(0.1)  # Allow consumer to recover
        await publisher.publish(message=message)
        message_body = await collect_future(self._test_future, timeout=10)

        # Assert
        assert consumer.channel_config.started
        assert message_body == message.body

    async def message_handler(self, message: IncomingMessage) -> None:
        self._test_future.set_result(message.body)



import asyncio
import logging

import pytest
from aio_pika import RobustConnection

from tornado_bunny import RabbitMQConnectionData, AsyncConnection


@pytest.mark.asyncio
async def test_async_connection_get_connection(event_loop: asyncio.AbstractEventLoop,
                                               logger: logging.Logger,
                                               rabbitmq_connection_data: RabbitMQConnectionData) -> None:
    # Arrange
    async_connection = AsyncConnection(rabbitmq_connection_data, logger, event_loop)

    # Act
    connection = await async_connection.get_connection()

    # Assert
    assert isinstance(connection, RobustConnection) and async_connection.is_connected()


@pytest.mark.asyncio
async def test_async_connection_connection_failure(event_loop: asyncio.AbstractEventLoop,
                                                   logger: logging.Logger,
                                                   rabbitmq_connection_data: RabbitMQConnectionData) -> None:
    # Arrange
    rabbitmq_connection_data.host = "1.1.1.1"
    async_connection = AsyncConnection(rabbitmq_connection_data,
                                       logger,
                                       event_loop,
                                       connection_attempts=2,
                                       timeout=0.1,
                                       attempt_backoff=1)

    # Act
    with pytest.raises(ConnectionError):
        await async_connection.get_connection()


@pytest.mark.asyncio
async def test_async_connection_close_connection(event_loop: asyncio.AbstractEventLoop,
                                                 logger: logging.Logger,
                                                 rabbitmq_connection_data: RabbitMQConnectionData) -> None:
    # Arrange
    async_connection = AsyncConnection(rabbitmq_connection_data, logger, event_loop)
    connection = await async_connection.get_connection()

    # Act
    await async_connection.close()

    # Assert
    assert not async_connection.is_connected()

    # Act
    await async_connection.close()

    # Assert
    assert not async_connection.is_connected()

    # Act
    new_connection = await async_connection.get_connection()

    # Assert
    assert isinstance(new_connection, RobustConnection) and async_connection.is_connected()
    assert async_connection.is_connected() and connection != new_connection



import asyncio
from logging import Logger
from typing import Union

import aio_pika
import aiormq
from aio_pika import Message, RobustExchange

from . import ChannelConfiguration, AsyncConnection


class Publisher:
    """
    Responsible for publishing messages to a given exchange with a given routing key
    """
    channel_config: ChannelConfiguration

    _exchange_name: str
    _exchange_type: str
    _routing_key: str
    _durable: bool
    _auto_delete: bool

    def __init__(self, connection: AsyncConnection, logger: Logger, exchange_name: str,
                 loop: asyncio.AbstractEventLoop = None, exchange_type: str = "topic", routing_key: str = None,
                 durable: bool = False, auto_delete: bool = False, prefetch_count: int = 1,
                 channel_number: int = None, publisher_confirms: bool = True, on_return_raises: bool = True):
        """
        :param connection: AsyncConnection to pass to ChannelConfiguration
        :param logger: Logger
        :param loop: Loop
        :param exchange_name: Exchange name to send messages to
        :param exchange_type: Exchange type
        :param routing_key: Default routing key to send the message with
        :param durable: Exchange durability
        :param auto_delete: Whether or not exchange auto deletes
        :param prefetch_count: Prefetch count for ChannelConfiguration
        :param channel_number: Channel number for ChannelConfiguration
        :param publisher_confirms: Publisher confirms for ChannelConfiguration
        :param on_return_raises: On return raises for ChannelConfiguration
        """
        self._channel_config = ChannelConfiguration(
            connection,
            logger,
            loop,
            prefetch_count,
            channel_number,
            publisher_confirms,
            on_return_raises
        )

        self._logger = logger
        self._exchange_name = exchange_name
        self._exchange_type = exchange_type
        self._routing_key = routing_key
        self._durable = durable
        self._auto_delete = auto_delete
        self._exchange = None

    @property
    def logger(self) -> Logger:
        """
        :return: self._logger
        """
        return self._logger

    @property
    def channel_config(self) -> ChannelConfiguration:
        """
        :return: self._channel_config
        """
        return self._channel_config

    async def _prepare_publish(self) -> None:
        """
        Ensures the channel is started and declares the desired exchange
        """
        await self.channel_config.ensure_channel()
        self._exchange = await self.channel_config.declare_exchange(
            self._exchange_name,
            exchange_type=self._exchange_type,
            durable=self._durable,
            auto_delete=self._auto_delete
        )

    async def _publish_message(self,
                               exchange: RobustExchange,
                               message: Message,
                               routing_key: str,
                               mandatory: bool = True,
                               immediate: bool = False,
                               timeout: Union[int, float, None] = None) -> Union[Exception, None]:
        self.logger.info(f"Publishing message. exchange: {exchange}; routing_key: {routing_key}; message: {message}")
        try:
            result = await exchange.publish(
                message=message,
                routing_key=routing_key,
                mandatory=mandatory,
                immediate=immediate,
                timeout=timeout
            )
            if self.channel_config.publisher_confirms and not result:
                raise ValueError("Publisher confirm failed")
        except aiormq.exceptions.ChannelNotFoundEntity as exc:
            self.logger.error(f"Exchange {self._exchange} was not found, resetting channel")
            return exc
        except aio_pika.exceptions.DeliveryError as exc:
            self.logger.error(f"Message {message} was returned, resetting channel")
            return exc
        except ValueError as exc:
            self.logger.error("Publisher confirm failed")
            return exc

    async def publish(self,
                      message: Message,
                      routing_key: str = None,
                      mandatory: bool = True,
                      immediate: bool = False,
                      timeout: Union[int, float, None] = None) -> None:
        """
        Publishes the given message to the desired exchange with the desired routing key
        :param message: Message to publish
        :param routing_key: Routing key to send the message with, defaults to the publisher's
        :param mandatory: Whether or not the message is mandatory
        :param immediate: Whether or not the message should be immediate
        :param timeout: Publish timeout
        """
        await self._prepare_publish()
        publish_exception = await self._publish_message(
            exchange=self._exchange,
            message=message,
            routing_key=routing_key or self._routing_key,
            mandatory=mandatory,
            immediate=immediate,
            timeout=timeout
        )
        if publish_exception:
            await self.channel_config.close_channel(publish_exception)
            await self.publish(message, routing_key, mandatory, immediate, timeout)

    async def default_exchange_publish(self,
                                       message: Message,
                                       routing_key: str,
                                       mandatory: bool = True,
                                       immediate: bool = False,
                                       timeout: Union[int, float, None] = None) -> None:
        exchange = await self.channel_config.get_default_exchange()
        publish_exception = await self._publish_message(
            exchange=exchange,
            message=message,
            routing_key=routing_key,
            mandatory=mandatory,
            immediate=immediate,
            timeout=timeout
        )
        if publish_exception:
            await self.channel_config.close_channel(publish_exception)
            await self.default_exchange_publish(message, routing_key, mandatory, immediate, timeout)



from .rabbitmq_connection_data import RabbitMQConnectionData
from .async_connection import AsyncConnection
from .channel_configuration import ChannelConfiguration
from .intentional_close_channel_error import IntentionalCloseChannelError
from .publisher import Publisher
from .consumer import Consumer
from .async_adapter import AsyncAdapter

__all__ = [
    "RabbitMQConnectionData",
    "AsyncConnection",
    "ChannelConfiguration",
    "IntentionalCloseChannelError",
    "Publisher",
    "Consumer",
    "AsyncAdapter"
]



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md', "r") as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst', "r") as history_file:
    history = history_file.read()

with open("requirements.txt", "r") as requirements_file:
    requirements = requirements_file.read()

setup_requirements = ['pytest-runner', 'vcversioner']

test_requirements = ['pytest>=3', ]

setup(
    author="Oded Shimon",
    author_email='audreyr@example.com',
    vcversioner={'vcs_args': ['git', 'describe', '--tags', '--long']},
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="RabbitMQ connector library for Python that is fully integrated with the aio-pika framework",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='tornado_bunny',
    name='tornado_bunny',
    packages=find_packages(include=['tornado_bunny', 'tornado_bunny.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/odedshimon/tornado-bunny',
    zip_safe=False,
)



import asyncio
from logging import Logger
from typing import Union

from aio_pika import connect_robust, RobustConnection

from . import RabbitMQConnectionData


class AsyncConnection:
    """
    Responsible for the management of a connection to a RabbitMQ server
    """
    _logger: Logger
    _loop: asyncio.AbstractEventLoop

    _connection: Union[RobustConnection, None]
    _rabbitmq_url: RabbitMQConnectionData
    _properties: dict

    _timeout: int
    _connection_attempts: int
    _attempt_backoff: int

    _connection_lock: asyncio.Lock

    def __init__(self, rabbitmq_connection_data: RabbitMQConnectionData, logger: Logger,
                 loop: asyncio.AbstractEventLoop = None, properties: dict = None,  timeout: Union[int, float] = 10,
                 connection_attempts: int = 5, attempt_backoff: int = 5):
        """
        :param rabbitmq_connection_data: RabbitMQConnectionData instance containing desired connection credentials
        :param logger: Logger
        :param loop: Event loop to use. Defaults to current event loop if None is passed.
        :param properties: Connection properties
        :param timeout: Connection timeout
        :param connection_attempts: Connection retry attempts
        :param attempt_backoff: Time waited between connection attempts
        """
        self._logger = logger
        self._loop = loop or asyncio.get_event_loop()

        self._rabbitmq_connection_data = rabbitmq_connection_data
        self._properties = properties or dict()

        self._timeout = timeout
        self._connection_attempts = connection_attempts
        self._attempt_backoff = attempt_backoff

        self._connection_lock = asyncio.Lock()
        self._connection = None

    @property
    def logger(self) -> Logger:
        """
        :return: self._logger
        """
        return self._logger

    async def get_connection(self) -> RobustConnection:
        """
        Gets or creates a connection to a RabbitMQ server.
        If we already have an open connection, returns it, otherwise opens a new connection.
        Avoids race conditions using a Lock.
        If creating a new connection fails, stops the event loop and raises an exception.
        :return: Robust RabbitMQ connection
        """
        async with self._connection_lock:
            if not self.is_connected():
                try:
                    self._connection = await self._connect()
                except ConnectionError:
                    self.logger.error("Failed to connect to RabbitMQ, stopping loop.")
                    self._loop.stop()
                    raise
        return self._connection

    def is_connected(self) -> bool:
        """
        :return: Whether or not there is currently an open connection stored in this object
        """
        return not ((self._connection is None) or (self._connection.connection is None) or self._connection.is_closed)

    async def _connect(self) -> RobustConnection:
        """
        Attempts to create a robust connection to a RabbitMQ server.
        Contains a retry mechanism which attempts to connect a configurable amount of times, with a configurable timeout
        and configurable backoff between each connection attempt.
        :return: Created robust connection
        :raise: ConnectionError in case connecting fails
        """
        for attempt_num in range(1, self._connection_attempts + 1):
            uri = self._rabbitmq_connection_data.uri()
            self.logger.info(f"Connecting to RabbitMQ, URI: {uri}, Attempt: {attempt_num}")
            try:
                connection = await connect_robust(url=uri,
                                                  loop=self._loop,
                                                  timeout=self._timeout,
                                                  client_properties=self._properties)
                return connection
            except (asyncio.TimeoutError, ConnectionError):
                self.logger.error(f"Connection attempt {attempt_num} / {self._connection_attempts} failed")
                if attempt_num < self._connection_attempts:
                    self.logger.debug(f"Going to sleep for {self._attempt_backoff} seconds")
                    await asyncio.sleep(self._attempt_backoff)
            except BaseException:
                raise ConnectionError("Unexpected exception during connection attempt")
        raise ConnectionError(f"Failed to connect to RabbitMQ server {self._connection_attempts} times")

    async def close(self) -> None:
        """
        Closes the connection stored in this object
        """
        if not self.is_connected():
            return
        await self._connection.close()
